                  /*--.          .-"-. 
                 /   ._.        /  ` \
                 \_  (__\       \_v_/
                 //   \\        //   \\
                ((     ))      ((     ))
 --""---""----""---""--
                  |||            |||                                               
   Authors note:   |              |                                                
                                                                                   
   This file is part of work done by Emil Persson, aka Humus.                      
   You're free to use the code in any way you like, modified, unmodified           
   or cut'n'pasted into your own work. But for the good of the whole               
   programming community I will ask you to as long as possible and to the          
   extent you find suitable to follow these points:                                
    * If you use this file and it's contents unmodified, please leave this note.   
    * If you use a major part of the file in your app, please credit the author(s) 
      in one way or the other as you find suitable, for instance a small notice    
      in the readme file may do, or perhaps a link to the authors site.            
    * Share your work and ideas too as much as you can.                            
     _______                                                                       
   /` _____ `\;,    Humus                                                          
  (__(^===^)__)';,  emiper-8@student.luth.se                         ___           
    /  :::  \   ,;  ICQ #47010716                                  /^   ^\         
   |   :::   | ,;'                                                (     )        
 '._______.'`--oOo--(_)--oOo--*/

#include <afxwin.h>
#include <atlbase.h>
#include <afxtempl.h>

#include <GL/GL.h>
#include <GL/GLU.h>

#include "3D/Resolution.h"

#include "OpenGL/glExtensions.h"
#include "OpenGL/Texture.h"
#include "Common/CPU.h"
#include "OpenGL/Light.h"

#include "Poly.h"

#define VK_ENTER 13

/* ################################################################################################################### */

class OpenGLWindow : public CFrameWnd {
public:
    OpenGLWindow();

	void DrawFrame();
protected:
	HDC hdc;
	HGLRC hrc;

	CRect windowedRect;
	int Width, Height;
	int Bpp;
	int Refresh;
	bool fullScreen;
	bool captureMouse;
	bool windowMoved;
	int middleX, middleY;

	bool Key[256];
	bool leftButton;
	float mouseSensibility;
	bool invertMouse;
	int mouseX, mouseY;

    bool showFPS;
    __int64 cpuHz;

	Vertex position;
	float Wx, Wy, Wz;

	TextureHandler *texHandler;
	TextureID text, lightTex;

	Set <Poly *> polygons;
	Set <Light *> lights;


	BOOL PreCreateWindow(CREATESTRUCT &cs);
	int  OnCreate(LPCREATESTRUCT lpCreateStruct);
	void OnSize(UINT nType, int cx, int cy);
	void OnClose();

	void OnKeyDown(UINT nChar, UINT nRepCnt, UINT nFlags);
	void OnKeyUp  (UINT nChar, UINT nRepCnt, UINT nFlags);
	void OnSysKeyDown(UINT nChar, UINT nRepCnt, UINT nFlags);

	void OnLButtonDown(UINT nFlags, CPoint point);
	void OnLButtonUp  (UINT nFlags, CPoint point);
	void OnMouseMove  (UINT nFlags, CPoint point);

	void OnKillFocus(CWnd *pNewWnd);

	DECLARE_MESSAGE_MAP ()
};

BEGIN_MESSAGE_MAP(OpenGLWindow, CFrameWnd)
	
	ON_WM_CREATE ()
	ON_WM_CLOSE ()

	ON_WM_KEYDOWN ()
	ON_WM_KEYUP ()
	ON_WM_SYSKEYDOWN ()

	ON_WM_LBUTTONDOWN ()
	ON_WM_LBUTTONUP ()
	ON_WM_MOUSEMOVE ()

	ON_WM_KILLFOCUS ()

END_MESSAGE_MAP()


/* ################################################################################################################### */

void OpenGLWindow::DrawFrame(){
    static bool first = true;

	int i, j;
    static __int64 prev, curr;
    static float currTime;
    float time, speed;


    prev = curr;
    curr = getCycleNumber();
    if (first){
        prev = curr;
        first = false;
    }
    time = (float) (100000.0 * double(curr - prev) / cpuHz);
    currTime += time;
    speed = time * 0.005f;


	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(75, (GLdouble) Width / (GLdouble) Height, 1, 32000);


    glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	glRotatef(Wz, 0.0f, 0.0f, 1.0f);
    glRotatef(Wx, 1.0f, 0.0f, 0.0f);
    glRotatef(Wy, 0.0f, 1.0f, 0.0f);

	glScalef(1,1,-1);
    glTranslatef(-position.x, -position.y, -position.z);

    float matrix[16];
    glGetFloatv(GL_MODELVIEW_MATRIX, matrix);
    Vertex dx(matrix[0], matrix[4], matrix[8]);
    Vertex dy(matrix[1], matrix[5], matrix[9]);
    Vertex dz(matrix[2], matrix[6], matrix[10]);


    Vertex dir(0,0,0);
	if (Key[VK_LEFT ]) dir -= dx;
	if (Key[VK_RIGHT]) dir += dx;
	
	if (Key[VK_UP  ]) dir -= dz;
	if (Key[VK_DOWN]) dir += dz;
	
	if (Key[VK_SHIFT  ]) dir -= dy;
	if (Key[VK_CONTROL]) dir += dy;
	
	if (Key['1']) Wz += 1.0f;
	if (Key['2']) Wz -= 1.0f;
	
	if (Key[VK_ENTER]){
		position = Vertex(256,256,256);
		Wx = 45;
		Wy = 225;
		Wz = 0;
	}
	
	if (lengthSqr(dir) != 0){
		dir.normalize();
	}

    position += speed * dir;


	// Start of Carmacks reversed
	
	glStencilMask(0xFFFFFFFF);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);

	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LEQUAL);

	lights[0]->setPosition(Vertex(250 * cosf(currTime * 0.00001f),  50 * sinf(currTime * 0.0000324f) - 50, 250 * sinf(currTime * 0.00001f)));
	lights[1]->setPosition(Vertex(150 * sinf(currTime * 0.00002f), 150 * sinf(currTime * 0.0000324f) + 50, 350 * cosf(currTime * 0.0000073f)));


	texHandler->apply();
	glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);
	glDepthMask(GL_TRUE);
	for (i = 0; i < polygons.getSize(); i++){
		polygons[i]->drawDepth();
	}


	glDepthMask(GL_FALSE);

	glEnable(GL_STENCIL_TEST);

	for (i = 0; i < lights.getSize(); i++){
		glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);
		glStencilFunc(GL_ALWAYS, 0, 0);
		glDepthFunc(GL_LESS);  // <=== Important to use GL_LESS and not GL_LEQUAL, 
		                       //      otherwise occluders will get incorrectly shadowed
		
		// We should really use this, but since know that only the two polygon in the middle
		// of this simple demo will cast shadows will save fillrate by skipping the others
		/*for (j = 0; j < polygons.getSize(); j++){
			polygons[j]->drawShadowVolume(lights[i]);
		}*/

		polygons[0]->drawShadowVolume(lights[i]);
		polygons[1]->drawShadowVolume(lights[i]);


		glDepthFunc(GL_LEQUAL);
		
		glStencilFunc(GL_EQUAL, 0, 0xFFFFFFFF);
		//glStencilOp(GL_ZERO, GL_ZERO, GL_ZERO);
		glStencilOp(GL_ZERO, GL_KEEP, GL_KEEP);

		
		glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
		for (j = 0; j < polygons.getSize(); j++){
			polygons[j]->drawLit(texHandler, lights[i], i > 0);
		}
	}

	glDisable(GL_STENCIL_TEST);
	glDepthMask(GL_TRUE);

	for (i = 0; i < polygons.getSize(); i++){
		polygons[i]->drawBase(texHandler);
	}

	// End of Carmacks reversed





	texHandler->setBlending(GL_ONE, GL_ONE);
	texHandler->setTextures(lightTex);
	texHandler->apply();
	
	Vertex v,r;
	for (i = 0; i < lights.getSize(); i++){
		v = lights[i]->getPosition();
		glColor4fv((GLfloat *) &lights[i]->getColor());
		
		glBegin(GL_QUADS);
			r = v - 30 * dx - 30 * dy;
			glTexCoord2f(0,0);
			glVertex3fv((GLfloat *) &r);
		
			r = v + 30 * dx - 30 * dy;
			glTexCoord2f(1,0);
			glVertex3fv((GLfloat *) &r);
		
			r = v + 30 * dx + 30 * dy;
			glTexCoord2f(1,1);
			glVertex3fv((GLfloat *) &r);
		
			r = v - 30 * dx + 30 * dy;
			glTexCoord2f(0,1);
			glVertex3fv((GLfloat *) &r);
		glEnd();
	}

	if (showFPS){
		static float fps[8];
		char str[256];

		for (i = 0; i < 7; i++){
			fps[i] = fps[i+1];
		}

		fps[7] = 100000 / time;
		float sum = 0;
		for (i = 0; i < 8; i++) sum += fps[i];
		
		sprintf(str,"%d", (int) (sum / 8 + 0.5f));

		texHandler->setTextures(text);
        texHandler->setTexEnvs(TEXENV_REPLACE);
		texHandler->setBlending(GL_ONE, GL_ONE);
		texHandler->apply();
		
		glColor3f(1,1,1);

        glDrawText(str, 0.02f, 0.02f, 0.03f, 0.03f);
	}

	SwapBuffers(hdc);
}


OpenGLWindow::OpenGLWindow(){
	Width   = 1024;
	Height  = 768;
	Bpp     = 32;
	Refresh = 85;
	fullScreen = true;
	captureMouse = true;

	mouseSensibility = 0.1f;
	invertMouse = true;

    middleX = Width / 2;
	middleY = Height / 2;
	if (fullScreen) setResolution(Width, Height, Bpp, Refresh);

    SetCursorPos(middleX, middleY);

	showFPS = true;

	position = Vertex(256,256,256);

	Wx = 45;
	Wy = 225;
	Wz = 0;

	ZeroMemory(&Key, sizeof(Key));
	leftButton = false;

	CRect r(0, 0, Width, Height);
	windowedRect = r;

	Create(NULL, "OpenGL", WS_POPUP | WS_CLIPCHILDREN | WS_CLIPSIBLINGS, r);
}

BOOL OpenGLWindow::PreCreateWindow(CREATESTRUCT &cs){
	CFrameWnd::PreCreateWindow(cs);

	cs.dwExStyle &= ~WS_EX_CLIENTEDGE;

	return 1;
}


int OpenGLWindow::OnCreate(LPCREATESTRUCT lpCreateStruct){
	if (CFrameWnd::OnCreate(lpCreateStruct) == -1) return -1;

	ShowWindow(SW_SHOW);
	SetForegroundWindow();
	SetFocus();

	hdc = ::GetDC((HWND) *this);

    static PIXELFORMATDESCRIPTOR pfd = {
        sizeof (PIXELFORMATDESCRIPTOR),
			1,
			PFD_DRAW_TO_WINDOW |
			PFD_SUPPORT_OPENGL |
			PFD_DOUBLEBUFFER,
			PFD_TYPE_RGBA,
			Bpp,
			0, 0, 0, 0, 0, 0,
			0, 0,
			0, 0, 0, 0, 0,
			24,
			8,
			0,
			PFD_MAIN_PLANE,
			0,
			0, 0, 0
    };

    int nPixelFormat = ChoosePixelFormat(hdc, &pfd);
    SetPixelFormat(hdc, nPixelFormat, &pfd);

	hrc = wglCreateContext(hdc);
	wglMakeCurrent(hdc,hrc);
	
	ShowCursor(FALSE);

    initCPU();

    cpuHz = getHz();

	initExtensions(hdc);

    CString missing;
    if (!GL_ARB_multitexture)             missing += "GL_ARB_multitexture\n";
    if (!GL_ARB_texture_env_dot3)         missing += "GL_ARB_texture_env_dot3\n";
    if (!GL_ARB_texture_env_combine)      missing += "GL_ARB_texture_env_combine\n";
    if (!GL_EXT_texture_edge_clamp)       missing += "GL_EXT_texture_edge_clamp\n";
    if (!GL_EXT_texture3D)                missing += "GL_EXT_texture3D\n";
    if (!(GL_ARB_texture_env_crossbar || GL_NV_texture_env_combine4)) missing += "GL_ARB_texture_env_crossbar/GL_NV_texture_env_combine4\n";

    if (missing.GetLength() > 0){
        MessageBox("The following required OpenGL extensions isn't supported by the graphic card and driver you're using:\n\n" + missing +
                   "\nThis application will now exit.", "Missing extensions", MB_OK);
        texHandler = NULL;
        PostMessage(WM_CLOSE);
        return 0;
    }

	texHandler = new TextureHandler();

	texHandler->createVectorField(64);

	text     = texHandler->loadFromFile("..\\Textures\\Font.png");
	lightTex = texHandler->loadFromFile("..\\Textures\\Corona2.png");

	int wall     = texHandler->loadFromFile("..\\Textures\\Fieldstone.dds");
	int wallBump = texHandler->loadFromFile("..\\Textures\\FieldstoneBump.png");
	texHandler->setFilter(wallBump, GL_LINEAR_MIPMAP_NEAREST);
	texHandler->toNormalMap(wallBump);
	int wood     = texHandler->loadFromFile("..\\Textures\\wood.dds");
	int woodBump = texHandler->loadFromFile("..\\Textures\\WoodBump.png");
	texHandler->setFilter(woodBump, GL_LINEAR_MIPMAP_NEAREST);
	texHandler->toNormalMap(woodBump);

	texHandler->uploadAll();

	polygons.add(new Poly(Vertex( -50, -200,   50), Vertex( 150, -200,  250), Vertex( 150, -200, -150), wood, woodBump, 1));
	polygons.add(new Poly(Vertex(-150, -150,  150), Vertex( 150, -150,  150), Vertex(-150, -150, -150), wood, woodBump, 1));

	polygons.add(new Poly(Vertex(-512, -512,  512), Vertex( 512, -512,  512), Vertex(-512, -512, -512), wall, wallBump, 3));
	polygons.add(new Poly(Vertex(-512,  512,  512), Vertex( 512,  512,  512), Vertex(-512, -512,  512), wall, wallBump, 3));
	polygons.add(new Poly(Vertex(-512,  512, -512), Vertex( 512,  512, -512), Vertex(-512,  512,  512), wall, wallBump, 3));
	polygons.add(new Poly(Vertex( 512,  512,  512), Vertex( 512,  512, -512), Vertex( 512, -512,  512), wall, wallBump, 3));
	polygons.add(new Poly(Vertex( 512,  512, -512), Vertex(-512,  512, -512), Vertex( 512, -512, -512), wall, wallBump, 3));
	polygons.add(new Poly(Vertex(-512,  512, -512), Vertex(-512,  512,  512), Vertex(-512, -512, -512), wall, wallBump, 3));

	lights.add(new Light(Vertex(0,0,0), Color(1,1,1), 800));
	lights.add(new Light(Vertex(0,0,0), Color(1,0,0), 800));

	return 0;
}

void OpenGLWindow::OnSize(UINT nType, int cx, int cy){
	glViewport(0, 0, cx, cy);
}

void OpenGLWindow::OnClose(){
	int i, len;

	len = polygons.getSize();
	for (i = 0; i < len; i++){
		delete polygons[i];
	}

	len = lights.getSize();
	for (i = 0; i < len; i++){
		delete lights[i];
	}

    if (texHandler != NULL){
        texHandler->unloadAll();
        delete texHandler;
    }


	wglMakeCurrent(NULL, NULL);
	wglDeleteContext(hrc);
	::ReleaseDC((HWND) *this,hdc);

	if (fullScreen) resetResolution();

	ShowCursor(TRUE);
	DestroyWindow();
}


void OpenGLWindow::OnKeyDown(UINT nChar, UINT nRepCnt, UINT nFlags){
	Key[nChar & 0xFF] = true;

    switch(nChar){
        case ' ':
            showFPS = !showFPS;
            break;
        case 'M':
            invertMouse = !invertMouse;
            break;
        case VK_ESCAPE:
            if (!fullScreen && captureMouse){
                ShowCursor(TRUE);
                captureMouse = false;
            } else {
                PostMessage(WM_CLOSE, 0,0);
            }
            break;
    }
}

void OpenGLWindow::OnKeyUp(UINT nChar, UINT nRepCnt, UINT nFlags){
	Key[nChar & 0xFF] = false;
}

void OpenGLWindow::OnSysKeyDown(UINT nChar, UINT nRepCnt, UINT nFlags){
	if (((nFlags >> 13) & 0x3) == 1 && nChar == VK_ENTER){
		if (fullScreen){
			resetResolution();
			middleX = (windowedRect.left + windowedRect.right)  / 2;
			middleY = (windowedRect.top  + windowedRect.bottom) / 2;
			MoveWindow(windowedRect, TRUE);

			fullScreen = false;
		} else {
			middleX = Width  / 2;
			middleY = Height / 2;
			MoveWindow(0, 0, Width, Height, TRUE);
			setResolution(Width, Height, Bpp, Refresh);

			captureMouse = true;
			ShowCursor(FALSE);

			fullScreen = true;
		}
	}
}

void OpenGLWindow::OnLButtonDown(UINT nFlags, CPoint point){
	mouseX = point.x;
	mouseY = point.y;

	windowMoved = false;
	leftButton = true;
}

void OpenGLWindow::OnLButtonUp(UINT nFlags, CPoint point){
	if (!captureMouse){
		if (!windowMoved){
			captureMouse = true;
			ShowCursor(FALSE);
		}
	}

	leftButton = false;
}

void OpenGLWindow::OnMouseMove(UINT nFlags, CPoint point){
	static bool changeMouse = false;

	if (captureMouse){
		ClientToScreen(&point);

		Wx += mouseSensibility * (point.y - middleY) * (invertMouse? -1 : 1);
		Wy += mouseSensibility * (point.x - middleX);
		
		if (changeMouse = !changeMouse) SetCursorPos(middleX, middleY);
	} else {
		if (leftButton){
			windowedRect += CSize(point.x - mouseX, point.y - mouseY);
			middleX = (windowedRect.left + windowedRect.right)  / 2;
			middleY = (windowedRect.top  + windowedRect.bottom) / 2;

			MoveWindow(windowedRect, TRUE);
			windowMoved = true;
		}
	}
}

void OpenGLWindow::OnKillFocus(CWnd *pNewWnd){
	if (fullScreen){
		resetResolution();
		middleX = (windowedRect.left + windowedRect.right)  / 2;
		middleY = (windowedRect.top  + windowedRect.bottom) / 2;
		MoveWindow(windowedRect, TRUE);
		fullScreen = false;
		captureMouse = false;
		ShowCursor(TRUE);
	}
}

/* ################################################################################################################### */


class OpenGLApp : public CWinApp {
public:
    virtual BOOL InitInstance();
	BOOL OnIdle(LONG lCount);
};


BOOL OpenGLApp::InitInstance(){
	
#ifdef _DEBUG
/*	int flag = _CrtSetDbgFlag(_CRTDBG_REPORT_FLAG); // Get current flag
	flag |= _CRTDBG_LEAK_CHECK_DF; // Turn on leak-checking bit
	flag |= _CRTDBG_CHECK_ALWAYS_DF; // Turn on CrtCheckMemory 
	_CrtSetDbgFlag(flag); // Set flag to the new value
*/
#endif

	m_pMainWnd = new OpenGLWindow();

	return TRUE;
}

BOOL OpenGLApp::OnIdle(LONG lCount){
	
	((OpenGLWindow *) m_pMainWnd)->DrawFrame();

	return TRUE;
}

OpenGLApp openGLApp;


/* ########################################## End of file ############################################################ */
